/*
 * gpio_imx27_lib.c
 *
 * Copyright 2013 Stefan Mavrodiev <support@olimex.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/select.h>
#include <pthread.h>
#include <unistd.h>
#include <sched.h>

#include "gpio_imx27_lib.h"

#define PIO_DDIR 0x00
#define PIO_OCR1 0x04
#define PIO_OCR2 0x08
#define PIO_ICONFA1 0x0C
#define PIO_ICONFA2 0x10
#define PIO_ICONFB1 0x14
#define PIO_ICONFB2 0x18
#define PIO_DR 0x1C
#define PIO_GIUS 0x20
#define PIO_SSR 0x24
#define PIO_ICR1 0x28
#define PIO_ICR2 0x2C
#define PIO_IMR 0x30
#define PIO_ISR 0x34
#define PIO_GPR 0x38
#define PIO_SWR 0x3C
#define PIO_PUEN 0x40
#define PIO_PMASK 0x600

//unsigned int IMX27_PIO_BASE = 0;
static volatile long int *gpio_map = NULL;

int imx27_gpio_init(void)
{
	int fd;
	unsigned int addr_start, addr_offset;
	unsigned int PageSize, PageMask;

	fd = open("/dev/mem", O_RDWR | O_SYNC);
	if (fd < 0) {
		return SETUP_DEVMEM_FAIL;
	}

	PageSize = sysconf(_SC_PAGESIZE);
	printf("%d\n", PageSize);
	PageMask = (~(PageSize - 1));

	addr_start = IMX27_GPIO_IO_BASE & PageMask;
	addr_offset = IMX27_GPIO_IO_BASE & ~PageMask;
	printf("%x %x %x\n", IMX27_GPIO_IO_BASE, addr_start, addr_offset);

	gpio_map =
	    (void *)mmap(0, PageSize * 2, PROT_READ | PROT_WRITE, MAP_SHARED,
			 fd, addr_start);
	if (gpio_map == MAP_FAILED) {
		return SETUP_MMAP_FAIL;
	}

	/*IMX27_PIO_BASE = (unsigned int)gpio_map;
	IMX27_PIO_BASE += addr_offset;*/

	close(fd);

	return SETUP_OK;
}

int imx27_gpio_set_cfgpin(unsigned int pin, unsigned int val)
{

	void *regaddr;
	unsigned int ocr_reg, iconfa_reg, iconfb_reg;
	unsigned int cfg;
	unsigned int cfg_iconfa;
	unsigned int cfg_iconfb;
	unsigned int bank = GPIO_BANK(pin);
	unsigned int offset = GPIO_NUM(pin);
	unsigned int off;

	if (gpio_map == NULL)
		return -1;

	regaddr = (void *)gpio_map + (bank * 0x100);

	/* direction */
	cfg = *(uint32_t *) regaddr + PIO_DDIR;
	cfg &= ~(0x01 << offset);
	cfg |= val << offset;
	*(uint32_t *) (regaddr + PIO_DDIR) = cfg;

	/* OCR / ICR */
	if (offset < 16) {
		off = offset;
		ocr_reg = PIO_OCR1;
		iconfa_reg = PIO_ICONFA1;
		iconfb_reg = PIO_ICONFB1;
	} else {
		off = offset - 16;
		ocr_reg = PIO_OCR2;
		iconfa_reg = PIO_ICONFA2;
		iconfb_reg = PIO_ICONFB2;
	}

	if (val == OUTPUT) {
		cfg = *(uint32_t *) regaddr + ocr_reg;
		cfg_iconfa = *(uint32_t *) regaddr + iconfa_reg;
		cfg_iconfb = *(uint32_t *) regaddr + iconfb_reg;

		//cfg = 0x30000000;
		cfg &= ~(0x03 << (2 * off));
		cfg |= (0x03 << (off * 2));

		//cfg_iconfa = 0xcfffffff;
		cfg_iconfa &= ~(0x03 << (off * 2));
		//cfg_iconfb = 0xcfffffff;
		cfg_iconfb &= ~(0x03 << (off * 2));

		*(uint32_t *) (regaddr + ocr_reg) = cfg;
		*(uint32_t *) (regaddr + iconfa_reg) = cfg_iconfa;
		*(uint32_t *) (regaddr + iconfb_reg) = cfg_iconfb;
	} else {
		/* ICONF TBD */
	}

	/* GIUS */
	*(uint32_t *) (regaddr + PIO_GIUS) |= (0x01 << offset);
	/* PUEN */
	*(uint32_t *) (regaddr + PIO_PUEN) &= ~(0x01 << offset);

	return 0;
}

int imx27_gpio_get_cfgpin(unsigned int pin)
{
	void *regaddr;
	unsigned int cfg;
	unsigned int bank = GPIO_BANK(pin);
	unsigned int offset = GPIO_NUM(pin);

	if (gpio_map == NULL)
		return -1;
	regaddr = (void *)gpio_map + (bank * 0x100);
	cfg = *(uint32_t *) regaddr + PIO_DDIR;
	cfg >>= offset;
	return (cfg & 0x1);
}

int imx27_gpio_output(unsigned int pin, unsigned int val)
{
	void *regaddr;
	unsigned int bank = GPIO_BANK(pin);
	unsigned int num = GPIO_NUM(pin);

	if (gpio_map == NULL)
		return -1;

	regaddr = (void *)gpio_map + (bank * 0x100);

	if (val)
		*(uint32_t *) (regaddr + PIO_DR) |= (0x01 << num);
	else
		*(uint32_t *) (regaddr + PIO_DR) &= ~(0x01 << num);

	return 0;
}

int imx27_gpio_input(unsigned int pin)
{
	void *regaddr;
	unsigned int dat;
	unsigned int bank = GPIO_BANK(pin);
	unsigned int num = GPIO_NUM(pin);

	if (gpio_map == NULL)
		return -1;

	/* la bank c'est GPIO_X = X = [A:F] x 0x100 
	 */
	regaddr = (void *)gpio_map + (bank * 0x100);

	dat = *(uint32_t *) (regaddr + PIO_SSR);
	dat >>= num;

	return (dat & 0x1);
}

void imx27_gpio_cleanup(void)
{
	unsigned int PageSize;
	if (gpio_map == NULL)
		return;

	PageSize = sysconf(_SC_PAGESIZE);
	munmap((void *)gpio_map, PageSize * 2);
}
